"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const createDebug = require("debug");
const fs = require("fs");
const mkdirp = require("mkdirp");
const ncp_1 = require("ncp");
const path = require("path");
const readline = require("readline");
const rimraf = require("rimraf");
const hbs = require("handlebars");
const compiler = require("node-elm-compiler");
const elm_view_options_1 = require("./elm-view-options");
const import_compiled_views_error_1 = require("./import-compiled-views-error");
const view_params_1 = require("./view-params");
class ElmViewEngine {
    constructor(_options = new elm_view_options_1.default()) {
        this._options = _options;
        this.getViewRequests = [];
        this.requestsRecyclingPool = [];
        this.debug = createDebug("elm-view-engine");
        this.debug.log = console.debug.bind(console);
    }
    get options() {
        return this._options;
    }
    /**
     * Compiles the views in the views folder and generates a javascript module
     * Outputted in the Options.compilePath folder (defaults to views folder).
     */
    compile() {
        this.debug("Starting compilation...");
        this.lastCompileError = undefined;
        return Promise.all([
            this.readViewsDir(),
            this.readTemplate(),
            this.ensureTempDirStructure(),
        ])
            .then((result) => __awaiter(this, void 0, void 0, function* () {
            const depsCopy = this.copyDependenciesFromProject(result[2]);
            const elmCode = yield this.compileAndOutputTemplate(result[0], result[1]);
            const modulePath = yield this.saveMainElmFile(result[2], elmCode);
            // Waiting for dependencies to be copied before compiling
            yield depsCopy;
            const jsModulePath = yield this.compileElmModule(this._options, modulePath);
            this.debug("Compilation succeeded. Cleaning...");
            // Awaiting to avoid interruption by the end of the process
            yield this.cleanGenerated();
            this.isUpdated = true;
            if (!this.watcher) {
                this.watchCompiledViews();
            }
            this.debug("Cleaning done");
            return jsModulePath;
        }))
            .catch((err) => __awaiter(this, void 0, void 0, function* () {
            this.debug("Compilation failed: %s", err);
            this.lastCompileError = err;
            yield this.cleanGenerated();
            throw err;
        }));
    }
    /**
     * Returns true if compilation is needed before calling #getView().
     */
    needsCompilation() {
        return new Promise(resolve => {
            fs.stat(path.join(this._options.compilePath, ElmViewEngine.OUTPUT_JS_FILENAME), err => {
                if (err) {
                    return resolve(true);
                }
                return resolve(false);
            });
        });
    }
    /**
     * Produce the view requested if it is in the views folder.
     * @param name name of the requested view. MUST be defined!
     * @param context context object of the view. Must have properties matching the elm deserializer in the elm view.
     */
    getView(name, context) {
        return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
            if (this.lastCompileError) {
                return reject(new import_compiled_views_error_1.default(this.lastCompileError));
            }
            const needsCompilation = yield this.needsCompilation();
            if (needsCompilation) {
                return reject(new Error("Views need to be compiled before rendering them"));
            }
            if (!name || name === "") {
                return reject(new Error("If you pass no name, you get no view!"));
            }
            // Lazy-instantianting the worker
            if (!this.worker || this.isUpdated) {
                const jsModulePath = path.join(this._options.compilePath, ElmViewEngine.OUTPUT_JS_FILENAME);
                try {
                    delete require.cache[require.resolve(jsModulePath)];
                    this.worker = require(jsModulePath).Main.worker();
                }
                catch (error) {
                    this.lastCompileError = error;
                    return reject(new import_compiled_views_error_1.default(error));
                }
                finally {
                    this.isUpdated = false;
                    this.watchCompiledViews();
                }
            }
            const id = this.createRequestHandler(resolve, reject, this.worker.ports.receiveHtml);
            this.worker.ports.receiveHtml.subscribe(this.getViewRequests[id]);
            this.worker.ports.getView.send(new view_params_1.default(id, name, context));
        }));
    }
    // Compilation
    watchCompiledViews() {
        this.watcher = fs.watch(path.join(this._options.compilePath, ElmViewEngine.OUTPUT_JS_FILENAME), () => {
            this.isUpdated = true;
            this.lastCompileError = undefined;
        });
    }
    cleanGenerated() {
        return new Promise((resolve, reject) => {
            rimraf(ElmViewEngine.GENERATION_DIR_BASE_PATH + "*", err2 => {
                if (err2) {
                    return reject(new Error("Failed to clean generated files. Please do it manually: " + err2));
                }
                return resolve();
            });
        });
    }
    readViewsDir() {
        const options = this._options;
        return new Promise((resolve, reject) => {
            fs.readdir(options.viewsDirPath, (errReadViews, files) => {
                if (errReadViews) {
                    return reject(errReadViews);
                }
                return resolve(files);
            });
        });
    }
    readTemplate() {
        return new Promise((resolve, reject) => {
            fs.readFile(ElmViewEngine.TEMPLATE_PATH, "UTF-8", (errReadTpl, content) => {
                if (errReadTpl) {
                    return reject(errReadTpl);
                }
                return resolve(content);
            });
        });
    }
    compileAndOutputTemplate(files, templateContent) {
        return __awaiter(this, void 0, void 0, function* () {
            const modules = yield Promise.all(files
                .filter(f => f.endsWith(".elm"))
                .map(f => path.join(this._options.viewsDirPath, f))
                .map(this.getModuleName));
            hbs.registerHelper("lastPart", (str, sep) => {
                const parts = str.split(sep);
                return parts[parts.length - 1];
            });
            const template = hbs.compile(templateContent);
            return template({ modules });
        });
    }
    getModuleName(file) {
        return new Promise((resolve, reject) => {
            const reader = readline.createInterface(fs.createReadStream(file));
            reader.on("line", (line) => {
                reader.close();
                const parts = line.split(" ");
                if (parts.length < 4) {
                    return reject(new Error(`Impossible to extract module name from ${file}`));
                }
                return resolve(parts[1]);
            });
        });
    }
    ensureTempDirStructure() {
        return new Promise((resolve, reject) => {
            mkdirp(path.dirname(ElmViewEngine.GENERATION_DIR_BASE_PATH), err => {
                if (err) {
                    return reject(err);
                }
                if (this.lastCompileError) {
                    return reject(new Error("Engine is in a faulty state. Aborting dir creation"));
                }
                fs.mkdtemp(ElmViewEngine.GENERATION_DIR_BASE_PATH, (err2, dir) => {
                    if (err2) {
                        return reject(err);
                    }
                    return resolve(dir);
                });
            });
        });
    }
    copyDependenciesFromProject(generatedPath) {
        return new Promise(resolve => {
            ncp_1.ncp(path.join(this._options.projectRoot, "elm-stuff"), path.join(generatedPath, "elm-stuff"), () => {
                // Not handling the error as it just means
                // dependencies will be downloaded
                resolve();
            });
        });
    }
    saveMainElmFile(generatedProjectPath, elmCode) {
        const modulePath = path.join(generatedProjectPath, "Main.elm");
        return new Promise((resolve, reject) => {
            fs.writeFile(modulePath, elmCode, err => {
                if (err) {
                    return reject(err);
                }
                return resolve(modulePath);
            });
        });
    }
    compileElmModule(options, modulePath) {
        return __awaiter(this, void 0, void 0, function* () {
            const projectPath = path.dirname(modulePath);
            // Loading engine + project elm configs
            const configs = yield Promise.all([
                this.loadElmPackageConfig(options.projectRoot),
                this.loadElmPackageConfig(__dirname),
            ]);
            const projConfig = configs[0];
            const engineConfig = configs[1];
            // Merging dependencies in engine elm config
            engineConfig.dependencies = Object.assign({}, projConfig.dependencies, engineConfig.dependencies);
            // Merging sources in engine elm config after normalizing all the paths
            engineConfig["source-directories"].push(...projConfig["source-directories"].map((dep) => {
                // paths are relative to project root, so we make them absolute
                let sourcePath = dep.replace("..", path.join(options.projectRoot, ".."));
                // If some paths use ./ we replace that with project root
                if (sourcePath.startsWith("./")) {
                    sourcePath = sourcePath.replace(".", options.projectRoot);
                }
                return path.relative(projectPath, sourcePath);
            }));
            yield this.outputElmPackageConfig(projectPath, engineConfig);
            try {
                const jsCode = yield compiler.compileToString(modulePath, {
                    cwd: projectPath,
                    pathToMake: options.elmMake,
                    verbose: process.env.NODE_ENV === "development",
                    yes: true,
                });
                return this.outputElmJsModule(jsCode);
            }
            catch (err) {
                this.debug(err);
                throw err;
            }
        });
    }
    outputElmJsModule(elmCode) {
        return new Promise((resolve, reject) => {
            const outPath = path.join(this._options.compilePath, ElmViewEngine.OUTPUT_JS_FILENAME);
            fs.writeFile(outPath, elmCode, error => {
                if (error) {
                    return reject(error);
                }
                return resolve(outPath);
            });
        });
    }
    outputElmPackageConfig(outputDir, config) {
        return new Promise((resolve, reject) => {
            fs.writeFile(path.join(outputDir, "elm-package.json"), JSON.stringify(config), err => {
                if (err) {
                    return reject(err);
                }
                return resolve();
            });
        });
    }
    loadElmPackageConfig(elmRoot) {
        return new Promise((resolve, reject) => {
            fs.readFile(path.join(elmRoot, "elm-package.json"), "UTF-8", (err, content) => {
                if (err) {
                    return reject(err);
                }
                return resolve(JSON.parse(content));
            });
        });
    }
    // View rendering
    createRequestHandler(resolve, reject, port) {
        // Reusing ids from past requests
        const recycledId = this.requestsRecyclingPool.shift();
        const reqId = recycledId !== undefined ? recycledId : this.getViewRequests.length;
        // Callback to be called when the view result is ready
        const cb = (view) => {
            if (view.id !== reqId) {
                return;
            }
            // Putting the req id aside for reuse and voiding it in the requests array
            this.requestsRecyclingPool.push(reqId);
            this.getViewRequests[reqId] = undefined;
            port.unsubscribe(cb);
            // Returning the result to the caller through promise callbacks
            if (view.error) {
                return reject(new Error(view.error));
            }
            return resolve(view.html);
        };
        this.getViewRequests[reqId] = cb;
        return reqId;
    }
}
ElmViewEngine.GENERATION_DIR_BASE_PATH = path.join(__dirname, "generated");
ElmViewEngine.TEMPLATE_PATH = path.join(__dirname, "Main.elm.hbs");
ElmViewEngine.OUTPUT_JS_FILENAME = "views.compiled.js";
exports.default = ElmViewEngine;
//# sourceMappingURL=elm-view-engine.js.map